const log = console.log;
const LS = localStorage;

const DropMain = { 
    lastSelectedStage: '',
    highScores: [],

    nowShowingReadyForm: false,
    nowStage: 0,

    selectedDiff: 0, // 0 : Easy, 1 : Normal, 2 : Hard, 3 : ...
    
    nowSliding: false,
    nowFading: false,

    stages: [],
    adContext: '',

    playingWindow: '',

    // initialize drop main page.
    init() {
        DropMain.hook();
        DropMain.loadStages();
        DropMain.nowStage = DropMain.lastSelectedStage ? DropMain.lastSelectedStage * 1 : 0;
    },

    // event handling
    hook() { 
        window.addEventListener('message', function (e) {
            log(e);
            DropMain.gameOverReciever(e.data.score, e.data.rank);
        });
    },

    // load and apply stage lists.
    loadStages() {
        $.ajax({
            type: "GET",
            url: "stageList.json",
            success(data) {
                DropMain.stages = data;
                DropMain.loadDatas().then(DropMain.refreshSelectForm());
            }
        });
    },

    // Load and save datas.
    async loadDatas() {
        DropMain.lastSelectedStage = LS['lastSelectedStage'] * 1 ? LS['lastSelectedStage'] * 1 : 0;
        DropMain.highScores = LS['highscores'] ? JSON.parse(LS['highscores']) : new Array(DropMain.stages.length).fill().map((x, i) => {
                return {
                    "easy": {
                        "score": 0,
                        "rank": "F"
                    },
                    "normal": {
                        "score": 0,
                        "rank": "F"
                    },
                    "hard": {
                        "score": 0,
                        "rank": "F"
                    },
                    "other": {
                        "score": 0,
                        "rank": "F"
                    }
                } 
            });
        if (DropMain.lastSelectedStage) DropMain.nowStage = DropMain.lastSelectedStage * 1;
    },

    // refresh drop select form.
    refreshSelectForm() {
        if (!DropMain.nowSliding) {
            DropMain.nowSliding = true;

            LS['lastSelectedStage'] = `${DropMain.nowStage}`;

            const nowStage = DropMain.nowStage;
            const slideWidth = DropMain.stages.length * 100;
            const mainColor = DropMain.stages[nowStage].stMainColor.split(",");

            const selDiffStr = DropMain.selectedDiff == 0 ? "Easy" : DropMain.selectedDiff == 1 ? "Normal" : "Hard";
            const nowStageHighScore = `${DropMain.highScores[nowStage][selDiffStr.toLowerCase()].score}`.padStart(7, "0");
            const nowStageRank = DropMain.highScores[nowStage][selDiffStr.toLowerCase()].rank;

            $('.stage-slide-content').html('');
            DropMain.stages.forEach(stage => {
                $('.stage-slide-content').append(
                    `<img class="stage-img" src="${stage.stBgPath}" alt="stageSlideContent" title="${stage.stTitle}" data-stno="${stage.stNumber}">`
                );
            });

            $('.select-form-black-bg')
                .css(`box-shadow`, `inset 0px 0px 60em 20em rgb(${mainColor[0].trim()}, ${mainColor[1].trim()}, ${mainColor[2].trim()})`)
                .css(`transition`, `box-shadow .3s`);

            $('.stage-slide-content').css({
                'width': `${slideWidth}%`,
                'left': `${nowStage * -100}%`
            });

            $('.stage-img').css({
                'width': `${100 / (slideWidth / 100)}%`,
                'height': '100%',
                'object-fit': 'cover'
            });

            ['Title', 'Author', 'Copyright', 'EasyDiff', 'NormalDiff', 'HardDiff'].forEach(name => {
                $(`.stage-${name}`).html(`${DropMain.stages[nowStage][`st${name}`]}`.replaceAll("&&&", "<br>"));
            });

            $('.start-btn').css('box-shadow', `0px 0px 50px rgb(${DropMain.stages[nowStage].stMainColor})`);
            $('.stage-highscore').html(`<span class="rank">${nowStageRank}</span> ${nowStageHighScore} <span class="rank">${nowStageRank}</span>`);

            setTimeout(() => DropMain.nowSliding = false, 300);
            DropMain.initAudio();
        }
    },

    // initialize audio && visualizer
    initAudio() {
        try {
            /*
            * Copyright (c) 2021 by Nick Jones (https://codepen.io/nfj525/pen/rVBaab)
            * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
            * The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
            * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
            */

            const audioElement = document.getElementById('bgm');
            audioElement.volume = 0.2;

            const audioVisualizer = document.getElementById('advs');
            DropMain.adContext = DropMain.adContext ? DropMain.adContext : new AudioContext();
            
            audioElement.src = DropMain.stages[DropMain.nowStage].stMusicPath;
            audioElement.load();
            audioElement.play();

            // repeat
            $('#bgm').on('ended', function () {
                audioElement.currentTime = 0;
                audioElement.play();
            });

            
            // audio visualizer
            const mdElementResource = DropMain.adContext.createMediaElementSource(audioElement);
            const mdAnalyser = DropMain.adContext.createAnalyser();

            audioVisualizer.width = window.innerWidth;
            audioVisualizer.height = window.innerHeight;
            
            const audioVisualizerContext = audioVisualizer.getContext("2d");

            mdElementResource.connect(mdAnalyser);
            mdAnalyser.connect(DropMain.adContext.destination);
            mdAnalyser.fftSize = 256;

            const bufLength = mdAnalyser.frequencyBinCount;
            const dataArray = new Uint8Array(bufLength);

            const audioVisualizerWidth = audioVisualizer.width;
            const audioVisualizerHeight = audioVisualizer.height;

            const visBarWidth = (audioVisualizerWidth / bufLength) * 2.5;
            let visBarHeight;
            let x;
            
            // rendering frames
            function renderFrame() {
                requestAnimationFrame(renderFrame);
                x = 0;

                mdAnalyser.getByteFrequencyData(dataArray);

                audioVisualizerContext.clearRect(0, 0, audioVisualizerWidth, audioVisualizerHeight);

                audioVisualizerContext.fillStyle = "rgba(0, 0, 0, 0)";
                audioVisualizerContext.fillRect(0, 0, audioVisualizerWidth, audioVisualizerHeight);

                for (let i = 0; i < bufLength; i++) {
                    visBarHeight = dataArray[i] * 1.5;

                    audioVisualizerContext.fillStyle = "rgba(255, 255, 255, 1)";
                    audioVisualizerContext.fillRect(x, audioVisualizerHeight - visBarHeight, visBarWidth, visBarHeight);

                    x += visBarWidth + 1;
                }
            }

            audioElement.play();
            renderFrame();
        } catch (er) {
            // ignore..
        }
    },

    showReady() {
        if (!DropMain.nowFading) {
            DropMain.nowFading = true;
            if (DropMain.nowShowingReadyForm) {
                $('.ready-form-bg').fadeOut(function () { DropMain.nowFading = false; });
            } else {
                $('.ready-form-bg').css('display', 'flex').hide().fadeIn(function () { DropMain.nowFading = false; });
            }
    
            DropMain.nowShowingReadyForm = !DropMain.nowShowingReadyForm;
        }
    },

    playGame() {
        const nowStage = DropMain.stages[DropMain.nowStage];
        const selDiffStr = DropMain.selectedDiff == 0 ? "Easy" :
            DropMain.selectedDiff == 1 ? "Normal" : "Hard";

        if (nowStage[`st${selDiffStr}HtmlPath`]) {
            $('.warning').css('display', 'flex').hide().fadeIn();
            $('.ready-form-bg').fadeOut();

            document.getElementById('bgm').pause();
            DropMain.playingWindow = window.open(nowStage[`st${selDiffStr}HtmlPath`], '_blank', `dialog, width=${screen.width}, height=${screen.height}`);
            
        }
    },

    gameOverReciever(score, rank) {
        log(score, rank);
        const selDiffStr = DropMain.selectedDiff == 0 ? "Easy" : DropMain.selectedDiff == 1 ? "Normal" : "Hard";
        
        if (DropMain.highScores[DropMain.nowStage][selDiffStr.toLowerCase()].score < score) {
            DropMain.highScores[DropMain.nowStage][selDiffStr.toLowerCase()] = {
                "score": score,
                "rank": rank
            }

            LS['highscores'] = JSON.stringify(DropMain.highScores);
        }

        $('.warning').fadeOut();
        DropMain.refreshSelectForm();

        setTimeout(() => DropMain.playingWindow.close(), 10000);
    }
}


$(_ => DropMain.init());